from . import config
from pu4c.common.utils import rpc_func

# 点云可视化
@rpc_func
def cloud_viewer(filepath=None, num_features=4, transmat=None,          # read_points
    points=None, point_labels=None, ds_voxel_size=None,
    cloud_uniform_color=None, cloud_colormap=None,                      # pointcloud color
    boxes3d=None, 
    vis=None, show_axis=True, run=True, 
    rpc=False):
    """
    快速查看单帧点云，支持 pcd/bin/npy/pkl/txt，输入文件路径或 ndarray 数组
    输入点云及带标签的边界框，可用于三维目标检测可视化
    输入点云及点云标签，可用于三维语义分割可视化、体素中心点及体素标签可视化、对比处理前后点云
    Examples:
        pu4c.det3d.app.cloud_viewer(filepath="/datasets/KITTI/object/training/velodyne/000000.bin", num_features=4)
        pu4c.det3d.app.cloud_viewer(points, boxes3d=boxes3d, rpc=True)
        pu4c.det3d.app.cloud_viewer(points=points, boxes3d=boxes3d_with_label, cloud_uniform_color=[0.99,0.99,0.99])
        pu4c.det3d.app.cloud_viewer(points=points, point_labels=point_labels, cloud_colormap=colormap)
    Keys:
        -/=: 调整点云点的大小
    Args:
        cloud_uniform_color: 自定义纯色点云颜色，例如白色 [1,1,1]
        cloud_colormap: 点云标签颜色表
        boxes3d: (N, 7)[xyz,lwh,yaw] or (N, 8)[xyz,lwh,yaw,cls]
        show_axis: 是否绘制坐标轴，如果不绘制那么会自动调整观察视角
        rpc: False 本地执行，True 远程执行
    """
    import open3d as o3d
    import numpy as np
    from .utils import common_utils, open3d_utils
    
    if vis is None:
        vis = o3d.visualization.Visualizer()
        vis.create_window()
        vis.get_render_option().point_size = 1
        vis.get_render_option().background_color = np.zeros(3)
    if show_axis:
        axis_geometry = o3d.geometry.TriangleMesh.create_coordinate_frame(size=1.0, origin=[0, 0, 0])
        vis.add_geometry(axis_geometry)

    if filepath is not None:
        points = common_utils.read_points(filepath, num_features=num_features, transmat=transmat)
    if points is not None:
        cloud_geometry = open3d_utils.create_pointcloud_geometry(
            points, labels=point_labels, ds_voxel_size=ds_voxel_size, 
            uniform_color=cloud_uniform_color, colormap=cloud_colormap, 
            )
        vis.add_geometry(cloud_geometry)
    if boxes3d is not None:
        boxes3d_geometry = open3d_utils.create_boxes3d_geometry(boxes3d)
        vis.add_geometry(boxes3d_geometry)

    if run:
        vis.run()
        vis.destroy_window()
@rpc_func
def voxel_viewer(voxel_centers, voxel_size, voxel_labels=None, voxel_colormap=None, show_axis=True, rpc=False):
    """
    输入体素中心点，可用于体素可视化
    输入体素中心点及标签，可以于 OCC 可视化
    Examples:
        pu4c.det3d.app.voxel_viewer(voxel_centers=voxel_coords*voxel_size, voxel_size=voxel_size)
        pu4c.det3d.app.voxel_viewer(voxel_centers, voxel_size, voxel_labels=labels, voxel_colormap=colormap)
    """
    import open3d as o3d
    import numpy as np
    from .utils import open3d_utils
    
    vis = o3d.visualization.Visualizer()
    vis.create_window()
    vis.get_render_option().point_size = 5
    vis.get_render_option().background_color = [1, 1, 1]
    if voxel_centers is not None:
        voxels_geometry = open3d_utils.create_voxels_geometry(voxel_centers, voxel_size)
        vis.add_geometry(voxels_geometry)

    # 如有标签优先按标签着色，否则着纯色
    cloud_viewer(points=voxel_centers, point_labels=voxel_labels, 
        cloud_uniform_color=[0, 1, 0], cloud_colormap=voxel_colormap, 
        vis=vis, show_axis=show_axis, run=True,
        )
@rpc_func
def cloud_viewer_panels(points_list=None, point_labels_list=None, boxes3d_list=None, 
    cloud_uniform_color=None, cloud_colormap=None,                                  # pointcloud color
    run=True, show_axis=True, offset=None, 
    rpc=False):
    """
    Examples:
        pu4c.det3d.app.cloud_viewer_panels(points_list=[points1, points2], boxes3d_list=[boxes3d1, boxes3d2], offset=[180, 0, 0])
    Args:
        offset: 面板之间的间隔，open3d 窗口坐标系，右前上
    """
    import open3d as o3d
    import numpy as np
    vis = o3d.visualization.Visualizer()
    vis.create_window()
    vis.get_render_option().point_size = 1
    vis.get_render_option().background_color = np.zeros(3)
    if show_axis:
        axis_geometry = o3d.geometry.TriangleMesh.create_coordinate_frame(size=1.0, origin=[0, 0, 0])
        vis.add_geometry(axis_geometry)

    offset = np.array([100, 0, 0]) if offset is None else np.array(offset)
    for i in range(len(points_list)):
        points = points_list[i][:, :3] + (offset * i)
        point_labels = point_labels_list[i] if point_labels_list is not None else None
        boxes3d = boxes3d_list[i] if boxes3d_list is not None else None
        if boxes3d_list is not None:
            boxes3d[:, :3] += (offset * i)
        cloud_viewer(
            points=points, point_labels=point_labels, boxes3d=boxes3d,
            cloud_uniform_color=cloud_uniform_color, cloud_colormap=cloud_colormap,
            vis=vis, show_axis=False, run=False, 
            )

    vis.run()
    vis.destroy_window()
@rpc_func
def cloud_player(root=None, pattern="*", num_features=4, filepaths=None,
    points_list=None, boxes3d_list=None, 
    cloud_uniform_color=None, show_axis=True,
    start=0, step=10, 
    rpc=False):
    """
    点云播放器，支持播放点云目录与点云列表
    Examples:
        pu4c.det3d.app.cloud_player(root="/datasets/KITTI/object/training/velodyne/", num_features=4, pattern="*.bin")
        pu4c.det3d.app.cloud_player(filepaths=filepaths, num_features=5, boxes3d_list=boxes3d_list, cloud_uniform_color=[0.99, 0.99, 0.99], rpc=True)
    Keys:
        A/D: pre/next one frame
        W/S: pre/next step frame
    """
    from glob import glob
    from .utils import open3d_utils

    assert (root is not None) or (filepaths is not None) or (points_list is not None)
    if root is not None:
        filepaths = sorted(glob(f'{root}/{pattern}'))
    length = len(points_list) if filepaths is None else len(filepaths)

    def switch(vis, i):
        vis.clear_geometries()
        print_msg = f"frame {i}" if root is None else f"frame {i}: {filepaths[i]}"
        print(print_msg)
        cloud_viewer(
            filepath=None if filepaths is None else filepaths[i],
            points=None if points_list is None else points_list[i],
            boxes3d=None if boxes3d_list is None else boxes3d_list[i],
            num_features=num_features,
            cloud_uniform_color=cloud_uniform_color, 
            vis=vis, show_axis=show_axis, run=False, 
            )
        # vis.poll_events()
        vis.update_renderer()
    
    open3d_utils.playcloud(switch, length, start=start, step=step)


@rpc_func
def plot_tsne2d(features, labels, 
    x="x", y="y", title="T-SNE",
    rpc=False):
    """
    Args:
        features: (N, M), N 个归一化的样本，每个 M 维
        labels: (N, ) 聚类标签
    """
    from sklearn.manifold import TSNE
    import numpy as np
    import pandas as pd
    import seaborn
    import matplotlib.pyplot as plt

    if features.shape[1] > 2:
        tsne = TSNE(n_components=2, init='pca', random_state=0)
        features = tsne.fit_transform(features)

    df = pd.DataFrame({'x': features[:, 0], 'y': features[:, 1], 'label': labels})
    seaborn.scatterplot(
        data=df, x=x, y=y, hue=df.label, 
        palette=seaborn.color_palette("hls", len(np.unique(labels))),
    ).set(title=title)
    plt.show()
@rpc_func
def plot_umap(features, labels, 
    x="x", y="y", title="T-SNE",
    rpc=False):
    # 与 t-SNE 相比，它在保持数据全局结构方面更加出色，但更慢
    # see https://umap-learn.readthedocs.io/en/latest/auto_examples/plot_mnist_example.html
    import umap # pip install umap-learn
    import numpy as np
    import pandas as pd
    import seaborn
    import matplotlib.pyplot as plt

    if features.shape[1] > 2:
        features = umap.UMAP(random_state=0).fit_transform(features)

    df = pd.DataFrame({'x': features[:, 0], 'y': features[:, 1], 'label': labels})
    seaborn.scatterplot(
        data=df, x=x, y=y, hue=df.label, 
        palette=seaborn.color_palette("hls", len(np.unique(labels))),
    ).set(title=title)
    plt.show()

def det3d_test_data(root=config.test_data_root):
    from pu4c.common.utils import TestDataDB
    return TestDataDB(dbname="det3d_test_data", root=root)



import rpyc, pickle
import os
from pu4c.common.config import rpc_server_ip, rpc_server_port, cache_dir

def rpc_func(func):
    def wrapper(*args, **kwargs):
        if ('rpc' in kwargs) and kwargs['rpc']:
            kwargs['rpc'] = False
            conn = rpyc.connect(rpc_server_ip, rpc_server_port)
            remote_method = getattr(conn.root, func.__name__, None)
            if remote_method:
                serialized_rets = remote_method(pickle.dumps(args), pickle.dumps(kwargs))
                conn.close()
                return pickle.loads(serialized_rets)
            else:
                raise AttributeError(f"Remote object has no attribute '{func.__name__}'")
        else:
            return func(*args, **kwargs) # python 中会为无返回值的函数返回 None
    return wrapper


def read_pickle(filepath):
    with open(filepath, 'rb') as f:
       data = pickle.load(f)
    return data
def write_pickle(filepath, data):
    with open(filepath, "wb") as f:
        pickle.dump(data, f)
class TestDataDB:
    def __init__(self, dbname="pu4c_test_data", root=cache_dir):
        mainfile = dbname + '.pkl'
        mainpath = os.path.join(root, mainfile)
        if not os.path.exists(mainpath):
            write_pickle(mainpath, data={"keys_dict": {}})
        
        self.root = root
        self.mainfile = mainfile
        self.mainpath = mainpath
        self.filesize = 1 * 1024**3
        self.keys_dict = read_pickle(mainpath)["keys_dict"]
    def get(self, key, default=None):
        return read_pickle(os.path.join(self.root, self.keys_dict[key]))[key] if key in self.keys_dict else default # 如果某个测试需要一批多个数据，则将其打包作为数据库的一项
    def add(self, key, data):
        if key in self.keys_dict:
            filepath = os.path.join(self.root, self.keys_dict[key])
            filedata = read_pickle(filepath)
            filedata[key] = data
            write_pickle(filepath, filedata)
            print(f"update {key}, data at {filepath}")
            return

        maindata = read_pickle(self.mainpath)
        maindata["keys_dict"][key] = self.keys_dict[key] = self.mainfile
        maindata[key] = data
        write_pickle(self.mainpath, data=maindata)

        if os.path.getsize(self.mainpath) > self.filesize:
            # 主文件过大时则将主文件重命名为新文件，并更新 keys_dict
            # 新文件中不删除 keys_dict 则在不进行删改数据库的情况下，也可以视作当前时间的备份
            from datetime import datetime
            now = datetime.now()
            timestamp = f"-{now.year % 100}{now.month:02d}{now.day:02d}"
            newfile = self.mainfile[:-len('.pkl')] + timestamp + '.pkl'
            
            if os.system(f"cp {os.path.join(self.root, self.mainfile)} {os.path.join(self.root, newfile)}") != 0:
                raise Exception(f"create new file {newfile} failed")
            print(f"create new file {newfile}")
            # udpate key
            maindata["keys_dict"].update({key:newfile for key, val in maindata["keys_dict"].items() if val == self.mainfile})
            self.keys_dict = new_keys_dict = maindata["keys_dict"]
            write_pickle(self.mainpath, data={"keys_dict": new_keys_dict})
    def remove(self, key):
        assert key in self.keys_dict
        filepath = os.path.join(self.root, self.keys_dict[key])
        # remove key
        maindata = read_pickle(self.mainpath)
        maindata["keys_dict"].pop(key)
        self.keys_dict.pop(key)
        # remove data
        if filepath == self.mainpath:
            maindata.pop(key)
            write_pickle(self.mainpath, maindata)
        else:
            write_pickle(self.mainpath, maindata)
            filedata = read_pickle(filepath)
            filedata.pop(key)
            write_pickle(filepath, filedata)
        print(f"remove {key}, data at {filepath}")
    def rename(self, key, new_key):
        assert key in self.keys_dict
        filepath = os.path.join(self.root, self.keys_dict[key])
        # rename key
        maindata = read_pickle(self.mainpath)
        self.keys_dict[new_key] = maindata["keys_dict"][new_key] = maindata["keys_dict"][key]
        maindata["keys_dict"].pop(key)
        self.keys_dict.pop(key)
        # rename data
        if filepath == self.mainpath:
            maindata[new_key] = maindata[key]
            maindata.pop(key)
            write_pickle(self.mainpath, maindata)
        else:
            write_pickle(self.mainpath, maindata)
            filedata = read_pickle(filepath)
            filedata[new_key] = filedata[key]
            filedata.pop(key)
            write_pickle(filepath, filedata)
        print(f"rename {key} to {new_key}, data at {filepath}")



import open3d as o3d
import numpy as np
from pu4c.det3d.utils import common_utils, color_utils

def create_pointcloud_geometry(points, labels=None, ds_voxel_size=None, uniform_color=None, colormap=None):
    cloud = o3d.geometry.PointCloud()
    cloud.points = o3d.utility.Vector3dVector(points[:, :3])
    if ds_voxel_size is not None:
        cloud = cloud.voxel_down_sample(ds_voxel_size)
    if labels is not None:
        colormap = [[255, 255, 255], [255, 0, 0], [0, 255, 0], [0, 0, 255]] if colormap is None else colormap
        colors = np.array(colormap)[labels.astype(np.int64)]
        cloud.colors = o3d.utility.Vector3dVector(colors)
    elif uniform_color is not None: cloud.paint_uniform_color(uniform_color)
    
    return cloud
def create_boxes3d_geometry(boxes3d, uniform_color=[0, 0.99, 0], show_heading=True):
    """
    Args:
        boxes3d: (N, 7)[xyz,lwh,yaw] or (N, 8)[xyz,lwh,yaw,cls]
    """
    # 计算角点与线框
    N, C = boxes3d.shape
    corners = common_utils.boxes3d_to_corners(boxes3d) # (N, 8, 3)
    box3d_lines = np.array(common_utils.get_oriented_bounding_box_lines(head_cross_lines=show_heading)) # (12, 2)
    boxes3d_lines = box3d_lines[None, :].repeat(N, axis=0) # (N, 12, 2) 此时其中每个顶点的编号都是相对于该框的相对编号，范围 [0,7]
    offsets = np.arange(start=0, stop=N * 8, step=8)[:, None, None] # (N, 1, 1)
    boxes3d_lines = boxes3d_lines + offsets # 顶点编号加偏移得到唯一的顶点编号，以便与顶点相对应

    line_sets = o3d.geometry.LineSet()
    line_sets.points = o3d.open3d.utility.Vector3dVector(corners.reshape((-1, 3)))
    line_sets.lines = o3d.open3d.utility.Vector2iVector(boxes3d_lines.reshape((-1, 2)))
    if C == 7: 
        line_sets.paint_uniform_color(uniform_color)
    else:
        if np.max(boxes3d[:, 7]) < 25:
            colormap = color_utils.color_det_class25
        else:
            raise ValueError("Number of classes is bigger than number of colors")
        box3d_colors = np.array(colormap)[boxes3d[:, 7].astype(np.int64)]
        boxes3d_colors = box3d_colors[:, None, :].repeat(boxes3d_lines.shape[1], axis=1) # (N, 12, 3)
        line_sets.colors = o3d.utility.Vector3dVector(boxes3d_colors.reshape((-1, 3)))

    return line_sets
def create_voxels_geometry(voxel_centers, voxel_size, uniform_color=[0, 0, 0]):
    """
    批量添加边框，可视化体素就是每个中心点都画一个框
    """
    # 计算每个非空体素的边框 (N,7)[xyz,wlh,yaw]
    N = voxel_centers.shape[0]
    wlh = np.tile(voxel_size, (N, 1)) # voxel_size 的第 0 维重复 N 次，第 1 维重复 1 次
    yaw = np.zeros((N, 1))
    boxes3d = np.concatenate((voxel_centers, wlh, yaw), axis=1)

    return create_boxes3d_geometry(boxes3d, uniform_color=uniform_color, show_heading=False)

def playcloud(switch_func, length, start=0, step=10):
    def prev(vis):
        global g_idx
        g_idx = max(g_idx - 1, 0)
        switch_func(vis, g_idx)
    def next(vis):
        global g_idx
        g_idx = min(g_idx + 1, length-1)
        switch_func(vis, g_idx)
    def prev_n(vis):
        global g_idx
        g_idx = max(g_idx - step, 0)
        switch_func(vis, g_idx)
    def next_n(vis):
        global g_idx
        g_idx = min(g_idx + step, length-1)
        switch_func(vis, g_idx)

    vis = o3d.visualization.VisualizerWithKeyCallback()
    vis.create_window()
    vis.get_render_option().point_size = 1
    vis.get_render_option().background_color = np.zeros(3)

    vis.register_key_callback(ord('W'), prev_n)
    vis.register_key_callback(ord('S'), next_n)
    vis.register_key_callback(ord('A'), prev)
    vis.register_key_callback(ord('D'), next) # 按小写，但这里要填大写
    # vis.register_key_callback(ord(' '), next) # space

    global g_idx
    g_idx = start
    switch_func(vis, start)
    vis.run()
    vis.destroy_window()


from .config import rpc_server_ip, rpc_server_port

def start_rpc_server():
    import rpyc, pickle
    from rpyc.utils.server import ThreadedServer
    from pu4c.det3d.app import cloud_viewer, voxel_viewer, cloud_viewer_panels, cloud_player, plot_tsne2d, plot_umap
    class RPCService(rpyc.Service):
        def exposed_cloud_viewer(self, serialized_args, serialized_kwargs):
            args, kwargs = pickle.loads(serialized_args), pickle.loads(serialized_kwargs)
            return pickle.dumps(cloud_viewer(*args, **kwargs))
        def exposed_voxel_viewer(self, serialized_args, serialized_kwargs):
            args, kwargs = pickle.loads(serialized_args), pickle.loads(serialized_kwargs)
            return pickle.dumps(voxel_viewer(*args, **kwargs))
        def exposed_cloud_viewer_panels(self, serialized_args, serialized_kwargs):
            args, kwargs = pickle.loads(serialized_args), pickle.loads(serialized_kwargs)
            return pickle.dumps(cloud_viewer_panels(*args, **kwargs))
        def exposed_cloud_player(self, serialized_args, serialized_kwargs):
            args, kwargs = pickle.loads(serialized_args), pickle.loads(serialized_kwargs)
            return pickle.dumps(cloud_player(*args, **kwargs))

        def exposed_plot_tsne2d(self, serialized_args, serialized_kwargs):
            args, kwargs = pickle.loads(serialized_args), pickle.loads(serialized_kwargs)
            return pickle.dumps(plot_tsne2d(*args, **kwargs))
        def exposed_plot_umap(self, serialized_args, serialized_kwargs):
            args, kwargs = pickle.loads(serialized_args), pickle.loads(serialized_kwargs)
            return pickle.dumps(plot_umap(*args, **kwargs))

    server = ThreadedServer(RPCService, port=rpc_server_port, auto_register=True)
    server.start()

def deep_equal(var1, var2, tol=None, ignore_keys=[], verbose=True, has_complex_type=False):
    """
    比较两个复杂变量是否相等，支持 dict, list, ndarray 等类型的嵌套
    Args:
        ignore_keys: list, 忽略字典中的键
        tol: tuple, (atol, rtol) 数值比较的容忍度
        has_complex_type: 是否包含复杂数据类型，即非 Python 内置类型或 numpy 类型
    """
    import numpy as np
    def convert_type(var):
        import torch
        if isinstance(var, dict):
            for k in var.keys():
                var[k] = convert_type(var[k])
        if isinstance(var, (list, tuple)):
            for i in range(len(var)):
                var[i] = convert_type(var[i])
        # 如果是张量，转 ndarray
        if isinstance(var, torch.Tensor):
            var = var.detach().cpu().numpy()
        # 如果是类实例，则转成字典继续转换，常规用法定义的类都有该方法并可以通过 vars 转成字典
        if hasattr(var, '__dict__'):
            var = convert_type(vars(var))
        return var

    def deep_equal_with_reason(var1, var2, reason):
        # 比较数据类型
        if type(var1) != type(var2):
            return False, f"{reason}: 类型不一致"
        
        # 如果是字典，则比较键值对
        if isinstance(var1, dict):
            if len(ignore_keys) > 0:
                var1 = {k: v for k, v in var1.items() if k not in ignore_keys}
                var2 = {k: v for k, v in var2.items() if k not in ignore_keys}
            elif var1.keys() != var2.keys():
                return False, f"{reason}: 键值不匹配"
            
            for k in var1.keys():
                is_equal, reason = deep_equal_with_reason(var1[k], var2[k], reason)
                if not is_equal:
                    return False, f"['{k}']{reason}"
            return True, reason
        
        # 如果是列表或元组，则逐元素比较
        if isinstance(var1, (list, tuple)):
            if len(var1) != len(var2):
                return False, f"{reason}: 列表长度不相等"
            
            for i, (x, y) in enumerate(zip(var1, var2)):
                is_equal, reason = deep_equal_with_reason(x, y, reason)
                if not is_equal:
                    return False, f"[{i}]{reason}"
            return True, reason
        
        # 如果是 NumPy 数组，则根据数组内容选择不同的比较方法
        if isinstance(var1, np.ndarray):
            if var1.shape != var2.shape:
                return False, f"{reason}: 数组形状不相等"
            
            if var1.dtype.kind in ['i', 'f']:
                # 数值数组，使用 np.isclose 进行比较 np.isclose: if abs(var1-var2) <= atol + rtol*abs(b)
                mask = np.isclose(var1, var2, atol=tol[0], rtol=tol[1]) if tol is not None else np.isclose(var1, var2) # rtol 相对误差容忍度
                if not np.all(mask):
                    idx = np.argwhere(~mask)[0] # 所有不相等值的索引，这里只打印第一个不相等的索引
                    return False, f"{reason}: 数组值在索引 {tuple(idx)} 处不相等"
            else:
                # 非数值数组（如字符串数组），使用 np.array_equal 进行比较
                if not np.array_equal(var1, var2):
                    idx = np.argwhere(var1 != var2)[0]
                    return False, f"{reason}: 数组值在索引 {tuple(idx)} 处不相等"
            return True, reason

        # 其他类型，直接比较
        return (True, reason) if var1 == var2 else (False, f"{reason}: 值不相等")

    if has_complex_type:
        var1, var2 = convert_type(var1), convert_type(var2)
    is_equal, reason = deep_equal_with_reason(var1, var2, reason='')
    if not is_equal and verbose:
        print(f"reason: {reason}")
    return is_equal


import numpy as np
import os
import copy


def read_points(filepath, num_features=4, transmat=None):
    """
    Args:
        transmat: 4x4 变换矩阵，某些点云可能希望直接进行坐标变换
    """
    filetype = os.path.splitext(filepath)[-1]
    if filetype == ".bin":
        points = np.fromfile(filepath, dtype=np.float32).reshape(-1, num_features)
    elif filetype == ".pcd" or filetype == '.ply':
        import open3d as o3d
        points = np.asarray(o3d.io.read_point_cloud(filepath).points)
    elif filetype == ".npy":
        points = np.load(filepath)
    elif filetype == ".pkl" or filetype == ".gz": # '.pkl.gz'
        import pandas as pd
        points = pd.read_pickle(filepath).to_numpy()
    elif filetype == ".txt":
        points = np.loadtxt(filepath, dtype=np.float32).reshape(-1, num_features)
    else:
        raise TypeError("unsupport file type")

    if transmat is not None:
        points[:, :3] = (transmat[:3, :3] @ points[:, :3].T +  transmat[:3, [3]]).T

    return points

def transform_matrix(rotation_mat, translation, inverse: bool = False) -> np.ndarray:
    """
    返回变换矩阵或变换矩阵的逆，要求传入的旋转矩阵和平移向量必须是有实际意义的(即从 3x4 矩阵中拿到的)，而不能对任意矩阵求逆
    """
    tm = np.eye(4)

    if inverse:
        rot_inv = rotation_mat.T
        trans = np.transpose(-np.array(translation))
        tm[:3, :3] = rot_inv
        tm[:3, 3] = rot_inv.dot(trans)
    else:
        tm[:3, :3] = rotation_mat
        tm[:3, 3] = np.transpose(np.array(translation))

    return tm
def limit_period(val,
                 offset: float = 0.5,
                 period: float = np.pi):
    """Limit the value into a period for periodic function.
    周期函数的区间映射，映射到 [-offset*period, (1-offset) * period]，floor 取小于输入 x 的最大整数
    常用于目标检测中规范化偏航角，坐标变换后的偏航角可能为n*[0,2pi]，以 2pi 为周期，可以将其调整为[-pi,pi]    
    Args:
        val (np.ndarray or Tensor): The value to be converted.
        offset (float): Offset to set the value range. Defaults to 0.5.
        period (float): Period of the value. Defaults to np.pi.

    Returns:
        np.ndarray or Tensor: Value in the range of
        [-offset * period, (1-offset) * period].
    """
    limited_val = val - np.floor(val / period + offset) * period
    return limited_val

def boxes3d_to_corners(boxes3d):
    """
             4-------- 6
           /|         /|
          5 -------- 3 .
          | |        | |
          . 7 -------- 1
          |/         |/
          2 -------- 0
    """
    dims = len(boxes3d.shape)
    if dims == 1:
        boxes3d = boxes3d[np.newaxis, :]

    template = np.array([
        [-1, -1, -1], [1, -1, -1], [-1, 1, -1], [-1, -1, 1],
        [1, 1, 1], [-1, 1, 1], [1, -1, 1], [1, 1, -1],
    ]) / 2.0

    corners3d = boxes3d[:, None, 3:6].repeat(8, 1).reshape(-1, 8, 3) * template
    corners3d = rotate_points_along_z(corners3d, boxes3d[:, 6])
    corners3d += boxes3d[:, None, 0:3]

    return corners3d[0] if dims == 1 else corners3d
def corners_to_boxes3d(corners):
    corners = copy.deepcopy(corners)
    centers = np.mean(corners, axis=1)
    corners -= centers[:, np.newaxis, :]
    dims = np.vstack([
        np.linalg.norm(corners[:, 1] - corners[:, 0], axis=-1),
        np.linalg.norm(corners[:, 2] - corners[:, 0], axis=-1),
        np.linalg.norm(corners[:, 3] - corners[:, 0], axis=-1),
    ]).T
    yaw = np.arctan2(corners[:, 1, 1] - corners[:, 0, 1], corners[:, 1, 0] - corners[:, 0, 0])
    return np.hstack([centers, dims, yaw[:, np.newaxis]])

def rotate_points_along_z(points, angle):
    cosa = np.cos(angle)
    sina = np.sin(angle)
    zeros = np.zeros(points.shape[0])
    ones = np.ones(points.shape[0])
    rot_matrix = np.stack((
        cosa,  sina, zeros,
        -sina, cosa, zeros,
        zeros, zeros, ones
    ), axis=1).reshape(-1, 3, 3)
    points_rot = points[:, :, :3] @ rot_matrix
    return points_rot
def mask_points_and_boxes_outside_range(points, limit_range, boxes3d=None):
    point_mask = (points[:, 0] >= limit_range[0]) & (points[:, 0] <= limit_range[3]) \
           & (points[:, 1] >= limit_range[1]) & (points[:, 1] <= limit_range[4])
    box_mask = ((boxes3d[:, :3] >= limit_range[:3]) & (boxes3d[:, :3]  <= limit_range[3:6])).all(axis=-1) if boxes3d is not None else None

    return point_mask, box_mask
def get_oriented_bounding_box_corners(xyz, lwh, axis_angles):
    """
    轴角转旋转矩阵（暂只考虑偏航）来将其旋转为有向包围盒，计算盒子的 8 个角点，添加连线
    Locals:
        lines: (10, 2), 预定义的 14 条连线
              4 -------- 6
             /|         /|
            5 -------- 3 .
            | |        | |
            . 7 -------- 1          
            |/         |/       z |/ x  
            2 -------- 0      y - 0
    Returns:
        corners: (N, 8, 3)
    """
    x, y, z = xyz
    l, w, h = lwh
    roll, pitch, yaw = axis_angles
    xdif, ydif, zdif = l/2, w/2, h/2
    offsets = np.array([
        [-xdif,  xdif, -xdif, -xdif, xdif, -xdif,  xdif,  xdif],
        [-ydif, -ydif,  ydif, -ydif, ydif,  ydif, -ydif,  ydif],
        [-zdif, -zdif, -zdif,  zdif, zdif,  zdif,  zdif, -zdif],
    ])
    R_x = np.array([
        [ 1, 0            ,  0          ],
        [ 0, np.cos(roll) , -np.sin(roll)],
        [ 0, np.sin(roll) ,  np.cos(roll)],
    ])
    R_y = np.array([
        [ np.cos(pitch),  0,  np.sin(pitch)],
        [ 0            ,  1,  0            ],
        [-np.sin(pitch),  0,  np.cos(pitch)],
    ])
    R_z = np.array([
        [ np.cos(yaw), -np.sin(yaw),  0],
        [ np.sin(yaw),  np.cos(yaw),  0],
        [ 0          ,  0          ,  1],
    ])
    R = R_x @ R_y @ R_z
    corners = (R @ offsets + np.array([[x], [y], [z]])).T
    
    return corners
def get_oriented_bounding_box_lines(head_cross_lines=True):
    lines = [
                [0, 2], [0, 3], [2, 5], [3, 5],
                [0, 1], [3, 6], [5, 4], [2, 7],
                [1, 6], [1, 7], [7, 4], [4, 6],
            ]
    if head_cross_lines:
        lines.extend([[1, 4], [6, 7]])
    return lines




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pu4c",
    version="1.1.6",
    packages=find_packages(exclude=["test"]),
    author="city945",
    author_email="city945@njust.edu.cn",
    url="https://github.com/city945",
    description="A python utils package for city945",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'rpyc',
    ],
)


import numpy as np

# 投影
def project_points_to_pixels(points, image_shape, transform_mat):
    """
    y = Rx 即 y(4,N) = transform_mat @ (4, N) 即 y(N,4) = (N,4) @ transform_mat.T
    Args:
        transform_mat: 4x4 激光雷达坐标系到像素坐标系的变换矩阵，它等于 相机内参矩阵 @ 激光雷达到相机坐标系的变换矩阵
    """
    points_hom = np.hstack((points[:, :3], np.ones((points.shape[0], 1), dtype=np.float32))) # [N, 4]
    points_pixel = (points_hom @ transform_mat.T)[:, :3]
    
    pixels_depth = points_pixel[:, 2]
    pixels = (points_pixel[:, :2].T / points_pixel[:, 2]).T # (N, 2)[col, row]

    # remove points outside the image
    mask = pixels_depth > 0
    mask = np.logical_and(mask, pixels[:, 0] > 0)
    mask = np.logical_and(mask, pixels[:, 0] < image_shape[1])
    mask = np.logical_and(mask, pixels[:, 1] > 0)
    mask = np.logical_and(mask, pixels[:, 1] < image_shape[0])

    return pixels, pixels_depth, mask
def project_pixels_to_points(pixels, depth, transform_mat):
    """
    Args:
        pixels: (N, 2)[x,y]
        depth: (N,)
    """
    N = depth.shape[0]
    points_cam = np.zeros((N, 3))
    points_cam[:, 2] = depth
    points_cam[:, :2] = pixels * depth[:, np.newaxis].repeat(2, axis=1)

    points_cam_hom = np.hstack((points_cam, np.ones((N, 1), dtype=np.float32))) # [N, 4]
    points_hom = points_cam_hom @ transform_mat.T
    return points_hom[:, :3]
def project_points_to_pixels_cv2(points, image_shape, lidar2cam_mat, intrinsics_4x4, dist_coeffs):
    """
    如果带畸变参数，可以调 cv2.projectPoints 进行处理，输入相机坐标系下的点，输出二维像素坐标
    Args:
        lidar2cam_mat: 4x4 激光雷达到相机坐标系的变换矩阵
        intrinsics_4x4: 相机内参矩阵
        dist_coeffs: 相机畸变参数
    """
    import cv2
    points_hom = np.hstack((points[:, :3], np.ones((points.shape[0], 1), dtype=np.float32))) # [N, 4]
    points_cam = points_hom @ lidar2cam_mat.T

    pixels_depth = points_cam[:, 2] # 像素深度即相机坐标系下的 z 
    rotation, translation = np.eye(3), np.zeros((3, 1))
    pixels, jac = cv2.projectPoints(points_cam[:, :3].T, rotation, translation, intrinsics_4x4[:3, :3], dist_coeffs)
    pixels = pixels.squeeze(axis=1)

    # remove points outside the image
    mask = pixels_depth > 0
    mask = np.logical_and(mask, pixels[:, 0] > 0)
    mask = np.logical_and(mask, pixels[:, 0] < image_shape[1])
    mask = np.logical_and(mask, pixels[:, 1] > 0)
    mask = np.logical_and(mask, pixels[:, 1] < image_shape[0])

    return pixels, pixels_depth, mask

def range_projection(points, height, width, fov):
    """
    lidar_to_rangeview 的仅线性拉伸简化版
    Args:
        fov: [fov_up, fov_down]
    """
    depth = np.linalg.norm(points[:, :3], ord=2, axis=1) # 按行求二范数，即距离
    yaw, pitch = -np.arctan2(points[:, 1], points[:, 0]), np.arcsin(points[:, 2] / depth)

    fov, fov_down = np.sum(np.abs(fov)), abs(fov[1])
    proj_x = 0.5 * (yaw / np.pi + 1.0)          # yaw=[-pi, pi] to [0.0, 1.0]
    proj_y = 1.0 - (pitch + fov_down) / fov     # pitch=[fov_up, fov_down] to [0.0, 1.0]
    proj_x *= width     # to [0.0, W]
    proj_y *= height    # to [0.0, H]

    # 坐标取整作为像素坐标
    proj_x = np.minimum(width - 1, np.floor(proj_x))
    proj_x = np.maximum(0, proj_x).astype(np.int32)  # to [0, W-1]
    proj_y = np.minimum(height - 1, np.floor(proj_y))
    proj_y = np.maximum(0, proj_y).astype(np.int32)  # to [0, H-1]

    range_image = np.full((height, width), -1, dtype=np.float32) # [H,W] range (-1 is no data)
    point_idx = np.full((height, width), -1, dtype=np.int32)     # [H,W] index (-1 is no data)
    range_image[proj_y, proj_x] = depth
    point_idx[proj_y, proj_x] = np.arange(depth.shape[0])

    return range_image, point_idx
def lidar_to_rangeview(points, height, width, max_depth=None, return_intensity=False,
                       fov=None,
                       resolution=None, fov_offset_down=None):
    """
    一般算法都是以线性拉伸的方式转 RV 图像，其实这就是一种柱坐标系体素栅格
    Examples:
        线性拉伸方式: lidar_to_rangeview(points, height, width, fov=fov)
        非线性拉伸方式: lidar_to_rangeview(points, height, width, resolution=resolution, fov_down=fov_down)
    Args:
        fov: [fov_up, fov_down, fov_left, fov_right] 增加水平视场角范围使可用于非机械雷达
        resolution: None 则以线性拉伸的方式，否则传入角度的分辨率 [res_y, res_x]，按实际填入（类似给定 voxel_size 的体素化）
        fov_offset_down: 对底部 fov 的偏移量，当可以刻意把显示范围扩大时，fov_offset_down 可以小于 fov_down 以使得画面居中
        return_intensity: 是否返回反射强度图像
    Returns: 
        range_image: 浮点数深度值，-1 表示无效点
        point_idx: 像素坐标到原始点云点坐标的索引，point = points[point_idx[y, x]] 或 point_idx.reshape(-1) rv_points = points[point_idx[point_idx != -1]]
        intensity_image: 反射强度，-1 表示无效点
    """
    depth = np.linalg.norm(points[:, :3], ord=2, axis=1) # 按行求二范数，即距离
    # 前左上坐标系，yaw 加负号反向一下以遵循从左到右递增的习惯
    yaw, pitch = -np.arctan2(points[:, 1], points[:, 0]), np.arcsin(points[:, 2] / depth)

    point_idx_raw = np.arange(points.shape[0])
    if resolution is None:
        assert (fov is not None)
        fov = fov if len(fov) == 4 else [fov[0], fov[1], -np.pi, np.pi]
        fov_height, fov_width = np.sum(np.abs(fov[:2])), np.sum(np.abs(fov[2:]))
        fov_up, fov_down, fov_left, fov_right = np.abs(fov)
        
        proj_x = yaw * width / fov_width + width / 2.0    # fov_width 线性拉伸到 width，并将 x 轴原点移动到图像中心列以符合前左上坐标系
        proj_y = -((pitch + fov_down) * height / fov_height) + height # + abs(fov_down) 将 patch 范围移动到 [0, +]

        proj_x = np.minimum(width - 1, np.floor(proj_x))
        proj_x = np.maximum(0, proj_x).astype(np.int32)  # to [0, W-1]
        proj_y = np.minimum(height - 1, np.floor(proj_y))
        proj_y = np.maximum(0, proj_y).astype(np.int32)  # to [0, H-1]
    else:
        assert (fov_offset_down is not None)
        (res_y, res_x), fov_offset_down = resolution, abs(fov_offset_down)

        proj_x = yaw / res_x + width / 2.0
        proj_y = -((pitch + fov_offset_down) / res_y) + height

        mask_in_x = np.logical_and(np.floor(proj_x) >=0, np.floor(proj_x) < width)
        mask_in_y = np.logical_and(np.floor(proj_y) >=0, np.floor(proj_y) < height)
        mask_in = np.logical_and(mask_in_x, mask_in_y)
        proj_x, proj_y = proj_x[mask_in].astype(np.int32), proj_y[mask_in].astype(np.int32)
        depth, points, point_idx_raw = depth[mask_in], points[mask_in], point_idx_raw[mask_in]

    range_image = np.full((height, width), -1, dtype=np.float32) # [H,W] range (-1 is no data)
    point_idx = np.full((height, width), -1, dtype=np.int32)     # [H,W] index (-1 is no data)
    # 按深度降序排序，则之后对投影到同一个像素点的多个激光点会自然取距离最近的点，类似最近回波模式
    indices = np.argsort(depth)[::-1]
    if max_depth is not None:
        depth[depth > max_depth] = -1 # 超出最远距离的点置为 -1 过滤掉
    proj_x, proj_y = proj_x[indices], proj_y[indices]
    range_image[proj_y, proj_x] = depth[indices]
    point_idx[proj_y, proj_x] = point_idx_raw[indices]

    if return_intensity:
        intensity_image = np.full((height, width), -1, dtype=np.float32)
        intensity_image[proj_y, proj_x] = points[indices][:, 3]
        return range_image, point_idx, intensity_image

    return range_image, point_idx
def rangeview_to_lidar(range_image, intensity_image=None, 
                       fov=None,
                       resolution=None, fov_offset_down=None):
    """
    modified from https://github.com/city945/LiDAR4D/blob/main/utils/convert.py
    Returns:
        points: (N,4) or (N,3)
    """
    height, width = range_image.shape
    proj_x, proj_y = np.meshgrid(np.arange(width, dtype=np.float32), np.arange(height, dtype=np.float32), indexing="xy")

    if resolution is None:
        assert (fov is not None)
        fov = fov if len(fov) == 4 else [fov[0], fov[1], -np.pi, np.pi]
        fov_height, fov_width = np.sum(np.abs(fov[:2])), np.sum(np.abs(fov[2:]))
        fov_up, fov_down, fov_left, fov_right = np.abs(fov)

        yaw = -((proj_x - width/2.0)*fov_width / width)
        pitch = (fov_up - proj_y*fov_height / height)
    else:
        assert (fov_offset_down is not None)
        (res_y, res_x), fov_offset_down = resolution, abs(fov_offset_down)

        yaw = -((proj_x - width/2.0) * res_x)
        pitch = ((height - proj_y) * res_y) - fov_offset_down

    dirs = np.stack([np.cos(pitch)*np.cos(yaw), np.cos(pitch)*np.sin(yaw), np.sin(pitch)], axis=-1)
    points = dirs * range_image.reshape(height, width, 1)
    if intensity_image is not None:
        points = np.concatenate([points, intensity_image.reshape(height, width, 1)], axis=2)

    num_features = 3 if intensity_image is None else 4
    points = points[np.where(range_image != -1)].reshape(-1, num_features)

    return points


# 采样
def farthest_point_sample(point, npoint):
    """
    Args:
        xyz: pointcloud data, [N, D]
        npoint: number of samples
    Return:
        centroids: sampled pointcloud index, [npoint, D]
    """
    N, D = point.shape
    xyz = point[:, :3]
    centroids = np.zeros((npoint,))
    distance = np.ones((N,)) * 1e10
    farthest = np.random.randint(0, N)  # N个点中随机取1个点作为初始采样点
    for i in range(npoint):
        centroids[i] = farthest         # 新增一个采样点，迭代 npoint 次
        centroid = xyz[farthest, :]     # 以采样点为中心点，计算到其他点的距离之和
        dist = np.sum((xyz - centroid) ** 2, -1) # axis=-1按最高维度雷达，二维则按列加，即行和即x**2+y**2+z**2
        mask = dist < distance          # 偏离太远的点置 false 丢弃
        distance[mask] = dist[mask]     # 只为为 true 的点更新有效距离， python 允许以bool数组做下标掩膜
        farthest = np.argmax(distance, -1)  # 取离当前点有效距离最远的点作为下一个采样点
    point = point[centroids.astype(np.int32)]
    return point
